#ifndef MYIO_H
#define MYIO_H

#include <cstdlib>
#include <iostream>
#include <string>
#include <fstream>
#include <omp.h>

using namespace std;

extern int gargc;
extern char **gargv;

template <typename T> T bswap(T val){
 T retVal;
 char *pVal = (char*) &val;
 char *pRetVal = (char*)&retVal;
 int size = sizeof(T);
 for(int i=0;i<size;i++){
     pRetVal[size-1-i] = pVal[i];
 }
 return retVal;
}
    
template <typename T> void bswap(T *y,T *x,int n){
 int i;
 #pragma omp parallel for default(shared) private(i)
 for(i=0;i<n;++i) y[i]=bswap(x[i]);
 return; 
}
    
void myio_init(int,char **);

void myio_close();

string get_s(const string &);

string get_s(const string &,const string &);

template <class T> T &close_file(T &ios){
 ios.close();
 ios.clear();
 return ios;
}

template <class T> T &open_file(T &ios,const string &file,ios_base::openmode mode=ios_base::in){
 ios.close();
 ios.clear();
 ios.open(file.c_str(),mode);
 return ios;
}

template <typename T> int get_param(const string &sa,T &a){
 string s=get_s(sa);
 if(s==" "){
  string filename=get_s("par");
  s=get_s(filename,sa);
 }
 if(s!=" "){
  a=stod(s);
  return 1;
 }
 else{
  cout<<"cannot find parameter "<<sa<<endl;
  return 0;
 }
}

template <typename T1,typename T2>
void get_param(const string &sa1,T1 &a1,const string &sa2,T2 &a2){
 get_param(sa1,a1);
 get_param(sa2,a2);
 return;
}

template <typename T1,typename T2,typename T3>
void get_param(const string &sa1,T1 &a1,const string &sa2,T2 &a2,const string &sa3,T3 &a3){
 get_param(sa1,a1);
 get_param(sa2,a2);
 get_param(sa3,a3);
 return;
}

template <typename T1> 
int to_header(const string &sh,const string &sa1,T1 a1){
 string filename=get_s(sh);
 if(filename==" "){
  cout<<"cannot find header "<<sh<<endl;
  return 0;
 }
 else{
  ofstream outfile;
  if(!open_file(outfile,filename,ofstream::app)){
   cout<<"cannot open header "<<sh<<endl;
   close_file(outfile);
   return 0;
  }
  else{
   outfile<<"\n"<<sa1<<"="<<a1<<"\n";
   close_file(outfile);
   return 1;
  }
 } 
}

template <typename T1,typename T2=int,typename T3> 
void to_header(const string &sh,const string &sa1,T1 a1,
                               const string &sa2,T2 a2){
 to_header(sh,sa1,a1);
 to_header(sh,sa2,a2);
 return;
}

template <typename T1,typename T2=int,typename T3> 
void to_header(const string &sh,const string &sa1,T1 a1,
                               const string &sa2,T2 a2,
                               const string &sa3,T3 a3){
 to_header(sh,sa1,a1);
 to_header(sh,sa2,a2);
 to_header(sh,sa3,a3);
 return;
}

template <typename T> int from_header(const string &sh,const string &sa,T &a){
 string filename=get_s(sh);
 if(filename==" "){
  cout<<"cannot find header "<<sh<<endl;
  return 0;
 }
 else{   
  string s=get_s(filename,sa);
  if(s!=" "){
   a=stod(s);
   return 1;
  } 
  else{
   cout<<"cannot find parameter "<<sa<<endl;
   return 0;
  }
 }
}

template <typename T1,typename T2> 
void from_header(const string &sh,const string &sa1,T1 &a1,const string &sa2,T2 &a2){
 from_header(sh,sa1,a1);
 from_header(sh,sa2,a2);
 return;
}

template <typename T1,typename T2,typename T3> 
void from_header(const string &sh,const string &sa1,T1 &a1,const string &sa2,T2 &a2,const string &sa3,T3 &a3){
 from_header(sh,sa1,a1);
 from_header(sh,sa2,a2);
 from_header(sh,sa3,a3);
 return;
}

template <typename T> 
void write(const string &s,const T *buff,size_t n,string format="native_float",ios_base::openmode mode=ios_base::out){
 string hfile=get_s(s),hfilename=hfile;
 size_t pos=hfile.rfind("/");
 if(pos!=string::npos) hfilename=hfile.substr(pos+1);
 if(hfile==" "){
  cout<<"cannot find header "<<s<<endl;
  return;
 }
 string path=get_s("/net/server2/homes/sep/huyle/.datapath","datapath");
 if(path==" "){
  cout<<"cannot find datapath"<<endl;
  return;
 }
 string bfile=path+hfilename+"@";
 ofstream out;
 if(!open_file(out,bfile,ofstream::binary|mode)){
  cout<<"cannot open binary file "<<bfile<<endl;
 }
 else{
  if(format=="native_float") out.write((char*)buff,sizeof(T)*n);  
  else{
   T x;
   for(size_t i=0;i<n;++i){
    x=bswap(buff[i]);
	out.write((char*)&x,sizeof(T));
   }
  }
 }
 close_file(out);
 if(!open_file(out,hfile,ofstream::app)){
  cout<<"cannot open header "<<hfile<<endl;
 }
 else{
  out<<"\n"<<"in="<<bfile<<"\n"<<"data_format="<<format<<endl;
 }
 close_file(out);
 return;
}

template <typename T> void read(const string &s,T *buff,size_t n,size_t pos=0){
 string hfile=get_s(s);
 if(hfile==" "){
  cout<<"cannot find header "<<s<<endl;
  return;
 }
 string bfile=get_s(hfile,"in");
 ifstream in;
 if(!open_file(in,bfile,ofstream::binary)){
  cout<<"cannot open binary file "<<bfile<<endl;
 }
 else{
  in.seekg(sizeof(T)*pos);
  string format=get_s(hfile,"data_format");
  if(format=="native_float") in.read((char*)buff,sizeof(T)*n);
  else{
   T x;
   for(size_t i=0;i<n;++i){
	in.read((char*)&x,sizeof(T));
	buff[i]=bswap(x);
   }
  }
  close_file(in);
 }
 return;
}

#endif
