#include <cstdlib>
#include <iostream>
#include <chrono>
#include <math.h>
#include <random>
#include "myio.h"
#include "mylib.h"
#include "dem.h"

using namespace std;

void reuss(double &c11,double &c33,double &c44,double &c12,double &c13,
            const double pa,const double a11,const double a33,const double a44,const double a12,const double a13,
            const double pb,const double b11,const double b33,const double b44,const double b12,const double b13){
 c33=1./(pa/a33+pb/b33);
 c44=0.;
 c11=1./(pa/a11+pb/b11);
 c13=1./(pa/a13+pb/b13);
 c12=1./(pa/a12+pb/b12);
 return;
}

void backus(double &c11,double &c33,double &c44,double &c12,double &c13,
            const double pa,const double a11,const double a33,const double a44,const double a12,const double a13,
            const double pb,const double b11,const double b33,const double b44,const double b12,const double b13){
 c33=1./(pa/a33+pb/b33);
 c44=1./(pa/a44+pb/b44);
 double c13c33=pa*a13/a33+pb*b13/b33;
 c13=c13c33*c33;
 double c132c33=pa*a13*a13/a33+pb*b13*b13/b33;
 double c11ab=pa*a11+pb*b11;
 c11=c13c33*c13c33*c33-c132c33+c11ab;
 c12=c11-c11ab+(pa*a12+pb*b12);
 return;
}

int main(int argc,char **argv){
 myio_init(argc,argv);
 double rho_water=1000.,vp_water=1500.;
 double c11_water=vp_water*vp_water*rho_water,c33_water=c11_water,c44_water=0.,c12_water=c11_water,c13_water=c11_water;
 double K_water=c11_water;
 
 double rho_quartz=2650.,vs_quartz=4000.,vp_quartz=6000.;
 double c44_quartz=vs_quartz*vs_quartz*rho_quartz,c11_quartz=vp_quartz*vp_quartz*rho_quartz;
 double c33_quartz=c11_quartz,c12_quartz=c11_quartz-2.*c44_quartz,c13_quartz=c12_quartz;
 double K_quartz=c12_quartz+2./3.*c44_quartz;

 int nx,nz;
 float dx,dz,ox,oz;
 get_param("nx",nx,"dx",dx,"ox",ox);
 get_param("nz",nz,"dz",dz,"oz",oz);

 float *Vshale=new float[nx*nz]();
 read("vshale",Vshale,nx*nz);
 float *phi_sand=new float[nx*nz]();
 read("sandporo",phi_sand,nx*nz);
 float *phi_shale=new float[nx*nz]();
 read("shaleporo",phi_shale,nx*nz);
 float *T=new float[nx*nz]();
 read("temp",T,nx*nz);

 float wdepth;
 get_param("wdepth",wdepth);
 int nwdepth=wdepth/dz+1;
 
 int nn;
 get_param("nn",nn);

 float *v=new float[nx*nz](); set(v,1500.*nn,nx*nwdepth);
 float *eps=new float[nx*nz](); set(eps,0.,nx*nwdepth);
 float *del=new float[nx*nz](); set(del,0.,nx*nwdepth);
 
 float *vv=new float[nx*nz*nn](); for(int i=0;i<nn;++i) set(vv+i*nx*nz,1500.,nx*nwdepth);
 float *ee=new float[nx*nz*nn]();
 float *dd=new float[nx*nz*nn]();
 float *Cvv=new float[nx*nz]();
 float *Cve=new float[nx*nz]();
 float *Cvd=new float[nx*nz]();
 float *Cee=new float[nx*nz]();
 float *Ced=new float[nx*nz]();
 float *Cdd=new float[nx*nz]();
 
// double rho_smectite=2400.,vp_smectite=2500.,vs_smectite=1500.,eps_smectite=0.1,del_smectite=-0.05,gam_smectite=0.18;
 double rho_smectite,vp_smectite,vs_smectite,eps_smectite,del_smectite,gam_smectite;
 get_param("rho_smectite",rho_smectite,"vp_smectite",vp_smectite,"vs_smectite",vs_smectite);
 get_param("eps_smectite",eps_smectite,"del_smectite",del_smectite,"gam_smectite",gam_smectite);
 fprintf(stderr,"smectite rho %f vp %f vs %f eps %f del %f gam %f\n",rho_smectite,vp_smectite,vs_smectite,eps_smectite,del_smectite,gam_smectite);

 double c33_smectite=vp_smectite*vp_smectite*rho_smectite;
 double c11_smectite=c33_smectite*(1.+2.*eps_smectite);
 double c44_smectite=vs_smectite*vs_smectite*rho_smectite;
 double c12_smectite=c11_smectite-2.*c44_smectite*(1.+2.*gam_smectite);
 double c13_smectite=sqrt((c33_smectite-c44_smectite)*((1.+2.*del_smectite)*c33_smectite-c44_smectite))-c44_smectite;
 double a1_smectite=c11_smectite+c33_smectite-2.*c13_smectite-4.*c44_smectite;
 double a2_smectite=c11_smectite-3.*c12_smectite+2.*c13_smectite-2.*c44_smectite;
 double a3_smectite=4.*c11_smectite-3.*c33_smectite-c13_smectite-2.*c44_smectite;
 double L_smectite=(c11_smectite+c33_smectite+5.*c12_smectite+8.*c13_smectite-4.*c44_smectite)/15.;
 double M_smectite=(7.*c11_smectite+2.*c33_smectite-5.*c12_smectite-4.*c13_smectite+12.*c44_smectite)/30.;
  
// double rho_illite=2600.,vp_illite=3500.,vs_illite=2500.,eps_illite=0.3,del_illite=0.05,gam_illite=0.28;
 double rho_illite,vp_illite,vs_illite,eps_illite,del_illite,gam_illite;
 get_param("rho_illite",rho_illite,"vp_illite",vp_illite,"vs_illite",vs_illite);
 get_param("eps_illite",eps_illite,"del_illite",del_illite,"gam_illite",gam_illite);
 fprintf(stderr,"illite rho %f vp %f vs %f eps %f del %f gam %f\n",rho_illite,vp_illite,vs_illite,eps_illite,del_illite,gam_illite);

 double c33_illite=vp_illite*vp_illite*rho_illite;
 double c11_illite=c33_illite*(1.+2.*eps_illite);
 double c44_illite=vs_illite*vs_illite*rho_illite;
 double c12_illite=c11_illite-2.*c44_illite*(1.+2.*gam_illite);
 double c13_illite=sqrt((c33_illite-c44_illite)*((1.+2.*del_illite)*c33_illite-c44_illite))-c44_illite;
 double a1_illite=c11_illite+c33_illite-2.*c13_illite-4.*c44_illite;
 double a2_illite=c11_illite-3.*c12_illite+2.*c13_illite-2.*c44_illite;
 double a3_illite=4.*c11_illite-3.*c33_illite-c13_illite-2.*c44_illite;
 double L_illite=(c11_illite+c33_illite+5.*c12_illite+8.*c13_illite-4.*c44_illite)/15.;
 double M_illite=(7.*c11_illite+2.*c33_illite-5.*c12_illite-4.*c13_illite+12.*c44_illite)/30.;
 
 float W200max=sqrt(2.5)*0.25/pi/pi,W400max=sqrt(4.5)*0.25/pi/pi;
 
 unsigned seed;
 get_param("seed",seed);
 if(seed==0) seed = chrono::system_clock::now().time_since_epoch().count(); 
 cout<<"seed is "<<seed<<endl;

 default_random_engine generator(seed);
 
 float bTt,eTt;
 get_param("bTt",bTt,"eTt",eTt);
 uniform_real_distribution<double> Tt_dist(bTt,eTt);
 fprintf(stderr,"bTt %f eTt %f\n",bTt,eTt);
 
 float bsigmat,esigmat;
 get_param("bsigmat",bsigmat,"esigmat",esigmat);
 uniform_real_distribution<double> sigmat_dist(bsigmat,esigmat);
 fprintf(stderr,"bsigmat %f esigmat %f\n",bsigmat,esigmat);
 
 float bphi0,ephi0;
 get_param("bphi0",bphi0,"ephi0",ephi0);
 uniform_real_distribution<double> phi0_dist(bphi0,ephi0);
 fprintf(stderr,"bphi0 %f ephi0 %f\n",bphi0,ephi0);
 
 float bmn,emn;
 get_param("bmn",bmn,"emn",emn);
 uniform_real_distribution<double> mn_dist(bmn,emn);
 fprintf(stderr,"bmn %f emn %f\n",bmn,emn);
 
 float ba,ea;
 get_param("ba",ba,"ea",ea);
 uniform_real_distribution<double> a_dist(ba,ea);
 fprintf(stderr,"ba %f ea %f\n",ba,ea);

 uniform_real_distribution<float> r_dist(0.,1.);

// chrono::high_resolution_clock::time_point start=chrono::high_resolution_clock::now();
 for(int j=0;j<nn;++j){ 
  double Tt=Tt_dist(generator);
  double sigmat=sigmat_dist(generator);
  double phi0=phi0_dist(generator);
  double m=mn_dist(generator);
  fprintf(stderr,"j %d Tt %f sigmat %f phi0 %f m %f\n",j,Tt,sigmat,phi0,m);  
   
  #pragma omp parallel for num_threads(16)
  for(int i=nx*nwdepth;i<nx*nz;++i){
//  for(int i=nx*nwdepth;i<nx*(nwdepth+40);++i){
   double c11,c33,c44,c12,c13;
   
   double Pillite=0.5+0.5*tanh((T[i]-Tt)*0.5/sigmat);
   double Vquartz=(1.-Vshale[i])*(1.-phi_sand[i]);
   double Vsmectite=Vshale[i]*(1.-phi_shale[i])*(1.-Pillite);
   double Villite=Vshale[i]*(1.-phi_shale[i])*Pillite;
   double phi=Vshale[i]*phi_shale[i]+(1.-Vshale[i])*phi_sand[i];
   
   if(phi<=phi0){
	double a;
    float r=r_dist(generator);
    if(r<0.5) a=a_dist(generator);
    else a=(1.-phi0)*phi/(1.-phi)/phi0;   
   
    double c11_sand,c33_sand,c44_sand,c12_sand,c13_sand;
    dem_aniso(c11_sand,c33_sand,c44_sand,c12_sand,c13_sand,c11_quartz,c33_quartz,c44_quartz,c12_quartz,c13_quartz,c11_water,c33_water,c44_water,c12_water,c13_water,a,phi_sand[i]);
 
    double W200=W200max*pow(1.-Vshale[i]*phi_shale[i]/phi0,m);
    double W400=W400max*pow(1.-Vshale[i]*phi_shale[i]/phi0,m);
  
    double c11a_smectite=L_smectite+2.*M_smectite+4.*sqrt(2.)/105.*pi*pi*(2.*sqrt(5.)*a3_smectite*W200+3.*a1_smectite*W400);
    double c33a_smectite=L_smectite+2.*M_smectite-16.*sqrt(2.)/105.*pi*pi*(sqrt(5.)*a3_smectite*W200-2.*a1_smectite*W400);
    double c44a_smectite=M_smectite-2.*sqrt(2.)/315.*pi*pi*(sqrt(5.)*(7.*a2_smectite+a3_smectite)*W200+24.*a1_smectite*W400);
    double c12a_smectite=L_smectite-4*sqrt(2)/315*pi*pi*(2*sqrt(5)*(7*a2_smectite-a3_smectite)*W200-3.*a1_smectite*W400);
    double c13a_smectite=L_smectite+4.*sqrt(2.)/315.*pi*pi*(sqrt(5.)*(7.*a2_smectite-a3_smectite)*W200-12.*a1_smectite*W400);
    
    double c11a_illite=L_illite+2.*M_illite+4.*sqrt(2.)/105.*pi*pi*(2.*sqrt(5.)*a3_illite*W200+3.*a1_illite*W400);
    double c33a_illite=L_illite+2.*M_illite-16.*sqrt(2.)/105.*pi*pi*(sqrt(5.)*a3_illite*W200-2.*a1_illite*W400);
    double c44a_illite=M_illite-2.*sqrt(2.)/315.*pi*pi*(sqrt(5.)*(7.*a2_illite+a3_illite)*W200+24.*a1_illite*W400);
    double c12a_illite=L_illite-4*sqrt(2)/315*pi*pi*(2*sqrt(5)*(7*a2_illite-a3_illite)*W200-3.*a1_illite*W400);
    double c13a_illite=L_illite+4.*sqrt(2.)/315.*pi*pi*(sqrt(5.)*(7.*a2_illite-a3_illite)*W200-12.*a1_illite*W400);
    
    double c11_clay,c33_clay,c44_clay,c12_clay,c13_clay;
    double c11_shale,c33_shale,c44_shale,c12_shale,c13_shale;
   
    backus(c11_clay,c33_clay,c44_clay,c12_clay,c13_clay,1.-Pillite,c11a_smectite,c33a_smectite,c44a_smectite,c12a_smectite,c13a_smectite,Pillite,c11a_illite,c33a_illite,c44a_illite,c12a_illite,c13a_illite);
    dem_aniso(c11_shale,c33_shale,c44_shale,c12_shale,c13_shale,c11_clay,c33_clay,c44_clay,c12_clay,c13_clay,c11_water,c33_water,c44_water,c12_water,c13_water,a,phi_shale[i]);
   
    backus(c11,c33,c44,c12,c13,1.-Vshale[i],c11_sand,c33_sand,c44_sand,c12_sand,c13_sand,Vshale[i],c11_shale,c33_shale,c44_shale,c12_shale,c13_shale);
   }   
   else{
    c11=1./(phi/c11_water+Vquartz/K_quartz+Vsmectite/c11_smectite+Villite/c11_illite);
    c33=1./(phi/c33_water+Vquartz/K_quartz+Vsmectite/c33_smectite+Villite/c33_illite);
	c44=0.;
    c12=1./(phi/c12_water+Vquartz/K_quartz+Vsmectite/c12_smectite+Villite/c12_illite);
    c13=1./(phi/c13_water+Vquartz/K_quartz+Vsmectite/c13_smectite+Villite/c13_illite);
    float cij=0.25*(c11+c33+c12+c13);
    c11=cij;c33=cij;c12=cij;c13=cij;
   }
   
   double rho=phi*rho_water+Vquartz*rho_quartz+Vsmectite*rho_smectite+Villite*rho_illite;
   
   vv[i+j*nx*nz]=sqrt(c33/rho);
   v[i]+=vv[i+j*nx*nz];
   ee[i+j*nx*nz]=(c11/c33-1.)*0.5;
   eps[i]+=ee[i+j*nx*nz];
   float temp=((c13+c44)*(c13+c44)-(c33-c44)*(c33-c44))*0.5/c33/(c33-c44)+0.1;
   dd[i+j*nx*nz]=min(ee[i+j*nx*nz],temp);
   del[i]+=dd[i+j*nx*nz];
  }
 }
 
// chrono::high_resolution_clock::time_point end=chrono::high_resolution_clock::now();
// chrono::duration<double> time=chrono::duration_cast<chrono::duration<double> >(end-start);
// cout<<"total time "<<time.count()<<" seconds"<<endl;
 
 scale(v,1./nn,nx*nz); 
 write("v",v,nx*nz);
 to_header("v","n1",nx,"d1",dx,"o1",0);
 to_header("v","n2",nz,"d2",dz,"o2",oz);
 
 write("vv",vv,nx*nz*nn);
 to_header("vv","n1",nx,"d1",dx,"o1",0);
 to_header("vv","n2",nz,"d2",dz,"o2",oz);
 to_header("vv","n3",nn,"d3",1.,"o3",0.);
 
 scale(eps,1./nn,nx*nz); 
 write("eps",eps,nx*nz);
 to_header("eps","n1",nx,"d1",dx,"o1",0);
 to_header("eps","n2",nz,"d2",dz,"o2",oz);

 write("ee",ee,nx*nz*nn);
 to_header("ee","n1",nx,"d1",dx,"o1",0);
 to_header("ee","n2",nz,"d2",dz,"o2",oz);
 to_header("ee","n3",nn,"d3",1.,"o3",0.);
 
 scale(del,1./nn,nx*nz); 
 write("del",del,nx*nz);
 to_header("del","n1",nx,"d1",dx,"o1",0);
 to_header("del","n2",nz,"d2",dz,"o2",oz);

 write("dd",dd,nx*nz*nn);
 to_header("dd","n1",nx,"d1",dx,"o1",0);
 to_header("dd","n2",nz,"d2",dz,"o2",oz);
 to_header("dd","n3",nn,"d3",1.,"o3",0.);
 
 for(int j=0;j<nn;++j){
  #pragma omp parallel for num_threads(16)
  for(int i=nwdepth*nx;i<nx*nz;++i){
   Cvv[i]+=(vv[i+j*nx*nz]-v[i])*(vv[i+j*nx*nz]-v[i]);
   Cve[i]+=(vv[i+j*nx*nz]-v[i])*(ee[i+j*nx*nz]-eps[i]);
   Cvd[i]+=(vv[i+j*nx*nz]-v[i])*(dd[i+j*nx*nz]-del[i]);
   Cee[i]+=(ee[i+j*nx*nz]-eps[i])*(ee[i+j*nx*nz]-eps[i]);
   Ced[i]+=(ee[i+j*nx*nz]-eps[i])*(dd[i+j*nx*nz]-del[i]);
   Cdd[i]+=(dd[i+j*nx*nz]-del[i])*(dd[i+j*nx*nz]-del[i]);
  }
 }
 
 scale(Cvv,1./nn,nx*nz); 
 write("Cvv",Cvv,nx*nz);
 to_header("Cvv","n1",nx,"d1",dx,"o1",0);
 to_header("Cvv","n2",nz,"d2",dz,"o2",oz);
 
 scale(Cve,1./nn,nx*nz); 
 write("Cve",Cve,nx*nz);
 to_header("Cve","n1",nx,"d1",dx,"o1",0);
 to_header("Cve","n2",nz,"d2",dz,"o2",oz);
 
 scale(Cvd,1./nn,nx*nz); 
 write("Cvd",Cvd,nx*nz);
 to_header("Cvd","n1",nx,"d1",dx,"o1",0);
 to_header("Cvd","n2",nz,"d2",dz,"o2",oz);
 
 scale(Cee,1./nn,nx*nz); 
 write("Cee",Cee,nx*nz);
 to_header("Cee","n1",nx,"d1",dx,"o1",0);
 to_header("Cee","n2",nz,"d2",dz,"o2",oz);
 
 scale(Ced,1./nn,nx*nz); 
 write("Ced",Ced,nx*nz);
 to_header("Ced","n1",nx,"d1",dx,"o1",0);
 to_header("Ced","n2",nz,"d2",dz,"o2",oz);
 
 scale(Cdd,1./nn,nx*nz); 
 write("Cdd",Cdd,nx*nz);
 to_header("Cdd","n1",nx,"d1",dx,"o1",0);
 to_header("Cdd","n2",nz,"d2",dz,"o2",oz);

 delete []Vshale;delete []phi_sand;delete []phi_shale;delete []T;delete []v;delete []eps;delete []del;

 delete []vv;delete []ee;delete []dd;
 delete []Cvv;delete []Cve;delete []Cvd;delete []Cee;delete []Ced;delete []Cdd;
 
 myio_close();
 return 0;
}

