#include <cstdio>
#include <cstdlib>
#include <cmath>
#include "myio.h"
#include "mylib.h"
#include "init.h"
#include "solvers.h"
#include "wave.h"

float source_zoom(bool &success,const float *c11,const float *c13,const float *c33,float &f,float *ggwavelet,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pwavelet,const float *d0,float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 float atr;
 
 float *wavelettr=new float[nt]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(::isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(wavelettr,1.,wavelet,atr,pwavelet,nt);
  
  ftr=sourceObjFuncGradient(ggwavelet,d0,c11,c13,c33,wavelettr,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggwavelet,pwavelet,nt);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
	success=true;
    memcpy(wavelet,wavelettr,nt*sizeof(float));
    
    delete []wavelettr;

    f=ftr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 success=false;
 memcpy(wavelet,wavelettr,nt*sizeof(float));
 
 delete []wavelettr;

 f=ftr;
 return atr;
}

float source_line_search(bool &success,const float *c11,const float *c13,const float *c33,float &f,float *ggwavelet,float f0,const float *gwavelet,const float *pwavelet,float a0,const float *d0,float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 float a;
 
 float g0=dot_product(gwavelet,pwavelet,nt);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew;

 float *twavelet=new float[nt]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(twavelet,1.,wavelet,anew,pwavelet,nt);
  
  fnew=sourceObjFuncGradient(ggwavelet,d0,c11,c13,c33,twavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
  
  if(::isnan(fnew) || ::isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggwavelet,pwavelet,nt);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=source_zoom(success,c11,c13,c33,f,ggwavelet,aold,anew,fold,fnew,gold,gnew,f0,g0,pwavelet,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
   delete []twavelet;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   success=true;
   a=anew;f=fnew;
   memcpy(wavelet,twavelet,nt*sizeof(float));
   delete []twavelet;
   return a;
  }
  
  if(gnew>=0.){
   a=source_zoom(success,c11,c13,c33,f,ggwavelet,anew,aold,fnew,fold,gnew,gold,f0,g0,pwavelet,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
   delete []twavelet;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);
 success=false;
 
 a=anew;f=fnew;
 
 memcpy(wavelet,twavelet,nt*sizeof(float));

 delete []twavelet;

 return a;
}

int main(int argc,char **argv){
 myio_init(argc,argv);

 int nnx,nnz,nx,nz,nt,npad;
 float ox,oz,ot,dx,dz,dt;
 init2d(nnx,nnz,nx,nz,nt,dx,dz,dt,ox,oz,ot,npad);

 float freq; get_param("freq",freq);
 float *wavelet=new float[nt]();
 if(freq>0.) ricker(wavelet,freq,nt,dt);
 if(freq<0.) read("wavelet",wavelet,nt);
 
 int ns;
 float os,ds,zs;
 init_shot(ns,ds,os,zs);
 int *sloc=new int[2*ns]();
 init_shot_loc(sloc,ns,ds,os,zs,dx,dz,ox,oz,npad);
 
 int nr;
 float o_r,dr,zr;
 init_rec(nr,dr,o_r,zr);
 int *rloc=new int[2*nr]();
 init_rec_loc(rloc,nr,dr,o_r,zr,dx,dz,ox,oz,npad);
 
 float rate; get_param("rate",rate);
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 float *d0=new float[nr*nnt*ns]();
 read("data",d0,nr*nnt*ns);
 
 float *taper=new float[nnx*nnz]();
 init_abc(taper,nx,nz,npad);

 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 
 init_model("c11",c11,nx,nz,npad); 
 init_model("c13",c13,nx,nz,npad); 
 init_model("c33",c33,nx,nz,npad); 
 
 int opt;
 get_param("opt",opt);

 if(opt==0){
  float *gwavelet=new float[nt]();
  
  float objfunc=sourceObjFuncGradient(gwavelet,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate); 
  fprintf(stderr,"obj func %f\n",objfunc);
  
  write("gwavelet",gwavelet,nt);
  to_header("gwavelet","n1",nt,"o1",0.,"d1",dt);
  
  delete []gwavelet;
 }
 else if(opt==1){
  float objfunc0=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
  fprintf(stderr,"initial obj func %f\n",objfunc0);
  
  float *gwavelet=new float[nt]();
  read("gwavelet",gwavelet,nt);
  
  float s0,a,b;
  get_param("s0",s0,"a",a,"b",b);
  
  float objfunc=objfunc0,s=(-1.)*s0;
  while(objfunc<=objfunc0){
   scale_add(wavelet,gwavelet,s,nt);
   objfunc=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
   fprintf(stderr,"steplength %f obj func %f\n",s,objfunc);
   s=a*s+b;
   memset(wavelet,0,nt*sizeof(float));
  }
 }
 else{
  float a0;
  get_param("a0",a0);
  
  int niter;
  get_param("niter",niter);
  
  float *f=new float[niter]();
  float *g=new float[niter]();
  float *gwavelet=new float[nt]();
  float *ggwavelet=new float[nt]();
  float *d=new float[nr*nnt*ns]();
  
  f[0]=sourceObjFuncGradient(gwavelet,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
  g[0]=dot_product(gwavelet,gwavelet,nt);

  float *pwavelet=new float[nt](); mynegate(pwavelet,gwavelet,nt);
  
  write("objfunc",f,1);
  to_header("objfunc","n1",1,"o1",0.,"d1",1.);
  
  write("wavelet",wavelet,nt);
  to_header("wavelet","n1",nt,"o1",0.,"d1",dt);
  to_header("wavelet","n2",1,"o2",0.,"d2",1.);
  
  for(int iter=1;iter<niter;++iter){
   fprintf(stderr,"iteration %d, obj func %f, normg2 %f\n",iter,f[iter-1],g[iter-1]);
 
   float pTg=dot_product(pwavelet,gwavelet,nt);

   if(pTg>=0.){
    fprintf(stderr,"Search direction is not a decent direction. Take steepest decent.\n");
    mynegate(pwavelet,gwavelet,nt);
   }
    
   bool success=true;
   float a=source_line_search(success,c11,c13,c33,f[iter],ggwavelet,f[iter-1],gwavelet,pwavelet,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
   
   if(success==false){
    memset(d,0,nr*nnt*ns*sizeof(float));
    synthetic_f(d,c11,c13,c33,pwavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
    float dTd=dot_product(d,d,nr*nnt*ns);
    a=g[iter-1]*nr*nnt*ns/dTd/nt;
    scale_add(wavelet,a,pwavelet,nt); 
    f[iter]=sourceObjFuncGradient(ggwavelet,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
   }

   a0=a;
   fprintf(stderr,"steplength %f\n",a);
   
   g[iter]=dot_product(ggwavelet,ggwavelet,nt);
 
   float temp=dot_product(ggwavelet,gwavelet,nt); 
   float bPR=(g[iter]-temp)/g[iter-1];
   float bFR=g[iter]/g[iter-1];
   float b;
   if(bPR<-bFR) b=-bFR;
   else if(bPR>bFR) b=bFR;
   else b=bPR;
   
   fprintf(stderr,"b %f\n",b);
 
   lin_comb(pwavelet,-1.,ggwavelet,b,pwavelet,nt); 
   
   float *pt;
   pt=gwavelet;gwavelet=ggwavelet;ggwavelet=pt;
   
   write("objfunc",f+iter,1,"native_float",std::ios_base::app);
   to_header("objfunc","n1",iter+1,"o1",0.,"d1",1.);
   write("wavelet",wavelet,nt,"native_float",std::ios_base::app);
   to_header("wavelet","n2",iter+1,"o2",0.,"d2",1.);
  }
  delete []f;delete []gwavelet;delete []ggwavelet;delete []pwavelet;
 }

 delete []wavelet;delete []d0;delete []sloc;delete []rloc;delete []taper;

 myio_close();
 return 0;
}
