#include <omp.h>
#include <cstdio>
#include "mylib.h"
#include "wave.h"
#include "kernels.h"

float sourceObjFuncGradient(float *gwavelet,const float *d0,const float *c11,const float *c13,const float *c33,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 double val=0;
 fprintf(stderr,"Calculate obj func and grad\n");
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 int nnx=nx+2*npad,nnz=nz+2*npad;
 float dx2=dx*dx,dz2=dz*dz,dt2=dt*dt;
 
 memset(gwavelet,0,nt*sizeof(float));
 
 int nGPUs;
 cudaGetDeviceCount(&nGPUs);
 //fprintf(stderr,"Total # GPUs = %d\n",nGPUs);

 float **d=new float*[nGPUs]();
 int **d_rloc=new int*[nGPUs]();
 float **d_c11=new float*[nGPUs]();
 float **d_c13=new float*[nGPUs]();
 float **d_c33=new float*[nGPUs]();
 float **d_taper=new float*[nGPUs]();
 float **d_d=new float*[nGPUs]();
 float **p0=new float*[nGPUs]();
 float **q0=new float*[nGPUs]();
 float **p1=new float*[nGPUs]();
 float **q1=new float*[nGPUs]();
 float **tp=new float*[nGPUs]();
 float **tq=new float*[nGPUs]();
 float **d_gwavelet=new float*[nGPUs]();
 float **d_d0=new float*[nGPUs]();
 float **d_d1=new float*[nGPUs]();
 float **d_d00=new float*[nGPUs]();
 float **d_d01=new float*[nGPUs]();
 float **tgwavelet=new float*[nGPUs]();
 
 for(int i=0;i<nGPUs;++i){
  cudaSetDevice(i);
  
  init_coeffs();
 
  d[i]=new float[nr*nnt]();
  
  cudaMalloc(&d_rloc[i],2*nr*sizeof(int));
  cudaMemcpy(d_rloc[i],rloc,2*nr*sizeof(int),cudaMemcpyHostToDevice);
   
  cudaMalloc(&d_c11[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_c13[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_c33[i],nnx*nnz*sizeof(float));
  cudaMemcpy(d_c11[i],c11,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_c13[i],c13,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_c33[i],c33,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  
  cudaMalloc(&d_taper[i],nnx*nnz*sizeof(float));
  cudaMemcpy(d_taper[i],taper,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
 
  cudaMalloc(&d_d[i],nr*sizeof(float));
 
  cudaMalloc(&p0[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&p1[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&q0[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&q1[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&tp[i],nnx*nnz*sizeof(float)); cudaMemset(tp[i],0,nnx*nnz*sizeof(float));
  cudaMalloc(&tq[i],nnx*nnz*sizeof(float)); cudaMemset(tq[i],0,nnx*nnz*sizeof(float));
  
  cudaMalloc(&d_gwavelet[i],nt*sizeof(float));
  
  cudaMalloc(&d_d0[i],nr*sizeof(float));
  cudaMalloc(&d_d1[i],nr*sizeof(float));
  cudaMalloc(&d_d00[i],nr*sizeof(float));
  cudaMalloc(&d_d01[i],nr*sizeof(float));
 
  tgwavelet[i]=new float[nnx*nnz]();
 }

 int npasses=(ns+nGPUs-1)/nGPUs;
 int shotLeft=ns;

 for(int pass=0;pass<npasses;++pass){
  int nGPUsNeed=min(shotLeft,nGPUs);
  //fprintf(stderr,"Pass %d, # GPUs %d\n",pass,nGPUsNeed);
  
  #pragma omp parallel for num_threads(nGPUsNeed)
  for(int i=0;i<nGPUsNeed;++i){
   cudaSetDevice(i);

   int is=pass*nGPUs+i;
   int slocxz=sloc[0+is*2]+sloc[1+is*2]*nnx;

   dim3 block(BLOCK_DIM_X,BLOCK_DIM_Y);
   dim3 grid((nnx-2*RADIUS+BLOCK_DIM_X-1)/BLOCK_DIM_X,(nnz-2*RADIUS+BLOCK_DIM_Y-1)/BLOCK_DIM_Y);

   cudaMemset(p0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(p1[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q1[i],0,nnx*nnz*sizeof(float));
 
   inject<<<1,1>>>(p1[i],q1[i],wavelet[0],slocxz,nnx,dt2);
  
   abc<<<grid,block>>>(p1[i],q1[i],d_taper[i],nnx,nnz);
  
   D<<<grid,block>>>(tp[i],tq[i],p1[i],q1[i],dx2,dz2,nnx,nnz);
  
   if(ratio==1){
    record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p1[i],q1[i],d_rloc[i],nr,nnx);
    cudaMemcpy(d[i]+nr,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
   }
   
   for(int it=2;it<nt;++it){
    //fprintf(stderr,"Time step it=%d\n",it);
  
    forwardC<<<grid,block>>>(p0[i],q0[i],p1[i],q1[i],tp[i],tq[i],d_c11[i],d_c13[i],d_c33[i],dt2,nnx,nnz);
  
    inject<<<1,1>>>(p0[i],q0[i],wavelet[it-1],slocxz,nnx,dt2);
  
    abc<<<grid,block>>>(p0[i],q0[i],d_taper[i],nnx,nnz);
    abc<<<grid,block>>>(p1[i],q1[i],d_taper[i],nnx,nnz);
    
    D<<<grid,block>>>(tp[i],tq[i],p0[i],q0[i],dx2,dz2,nnx,nnz);
  
    if(it%ratio==0){
     record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p0[i],q0[i],d_rloc[i],nr,nnx);
     cudaMemcpy(d[i]+(it/ratio)*nr,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
    }
  
    float *pt=p0[i]; 
    p0[i]=p1[i];
    p1[i]=pt;
    pt=q0[i];
    q0[i]=q1[i];
    q1[i]=pt;
   }
 
   cudaMemset(p0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(p1[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q1[i],0,nnx*nnz*sizeof(float));
  
   cudaMemset(d_gwavelet[i],0,nnx*nnz*sizeof(float));
   
   cudaMemcpy(d_d00[i],d0+(nt-1)/ratio*nr+is*nr*nnt,nr*sizeof(float),cudaMemcpyHostToDevice);
   cudaMemcpy(d_d0[i],d[i]+(nt-1)/ratio*nr,nr*sizeof(float),cudaMemcpyHostToDevice);
   inject<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(p0[i],q0[i],d_d00[i],d_d01[i],d_d0[i],d_d1[i],0.,d_rloc[i],nr,nnx,dt2);
  
   abc<<<grid,block>>>(p0[i],q0[i],d_taper[i],nnx,nnz);
  
   extract<<<1,1>>>(d_gwavelet[i]+nt-2,p0[i],q0[i],slocxz,nnx,dt2);
   
   for(int it=nt-3;it>=0;--it){
    //fprintf(stderr,"Time step it=%d\n",it);
    
    C<<<grid,block>>>(tp[i],tq[i],p0[i],q0[i],d_c11[i],d_c13[i],d_c33[i],nnx,nnz);
  
    backwardD<<<grid,block>>>(p1[i],q1[i],p0[i],q0[i],tp[i],tq[i],dx2,dz2,dt2,nnx,nnz);
  
    float f=(it+1.)/ratio;
    int i1=f;
    if((it+2)%ratio==0){
     cudaMemcpy(d_d1[i],d[i]+i1*nr,nr*sizeof(float),cudaMemcpyHostToDevice);
     cudaMemcpy(d_d01[i],d0+i1*nr+is*nr*nnt,nr*sizeof(float),cudaMemcpyHostToDevice);
     float *pt=d_d0[i]; 
     d_d0[i]=d_d1[i];
     d_d1[i]=pt;
     pt=d_d00[i];
     d_d00[i]=d_d01[i];
     d_d01[i]=pt;
    }
    f=f-i1;
    inject<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(p1[i],q1[i],d_d00[i],d_d01[i],d_d0[i],d_d1[i],f,d_rloc[i],nr,nnx,dt2);
  
    abc<<<grid,block>>>(p0[i],q0[i],d_taper[i],nnx,nnz);
    abc<<<grid,block>>>(p1[i],q1[i],d_taper[i],nnx,nnz);
  
    extract<<<1,1>>>(d_gwavelet[i]+it,p1[i],q1[i],slocxz,nnx,dt2);
  
    float *pt=p0[i]; 
    p0[i]=p1[i];
    p1[i]=pt;
    pt=q0[i];
    q0[i]=q1[i];
    q1[i]=pt;
   }
   cudaMemcpy(tgwavelet[i],d_gwavelet[i],nt*sizeof(float),cudaMemcpyDeviceToHost);
  }
  
  for(int i=0;i<nGPUsNeed;++i){
   int is=pass*nGPUs+i;
   #pragma omp parallel for num_threads(16) reduction(+:val)
   for(int irt=0;irt<nr*nnt;++irt){
    val+=(d[i][irt]-d0[irt+is*nr*nnt])*(d[i][irt]-d0[irt+is*nr*nnt]);
   }
   
   #pragma omp parallel for num_threads(16) shared(i)
   for(int it=0;it<nt;++it){
    gwavelet[it]+=tgwavelet[i][it];
   }
  }
  
  shotLeft-=nGPUsNeed;
 }

 for(int i=0;i<nGPUs;++i){
  cudaSetDevice(i);
  delete []d[i];
  cudaFree(d_rloc[i]);
  cudaFree(d_c11[i]);cudaFree(d_c13[i]);cudaFree(d_c33[i]); 
  cudaFree(d_taper[i]);
  cudaFree(d_d[i]);
  cudaFree(p0[i]);cudaFree(p1[i]);cudaFree(q0[i]);cudaFree(q1[i]);cudaFree(tp[i]);cudaFree(tq[i]);
  cudaFree(d_gwavelet[i]);
  cudaFree(d_d0[i]);cudaFree(d_d1[i]);cudaFree(d_d00[i]);cudaFree(d_d01[i]);
  delete []tgwavelet[i];
 }
 
 delete []d;
 delete []d_rloc;
 delete []d_c11;delete []d_c13;delete []d_c33;
 delete []d_taper;
 delete []d_d;
 delete []p0;delete []p1;delete []q0;delete []q1;delete []tp;delete []tq;
 delete []d_gwavelet;
 delete []d_d0;delete []d_d1;delete []d_d00;delete []d_d01;
 delete []tgwavelet;

 return 0.5*val;
}
